import time
from datetime import datetime, timezone
from queue import Queue
from threading import Event, Thread

import pandas
import serial

from naneos.partector2._data_structure import PARTECTOR2_DATA_STRUCTURE
from naneos.partector2._lambda_upload import _get_lambda_upload_list_serial


class Partector2(Thread):
    SERIAL_RETRIES = 5  # MAGIC, number of retries when command fails
    TIMEOUT_SER = 0.1  # MAGIC, timeout of serial connection
    TIMEOUT_INFO = TIMEOUT_SER + 0.05  # MAGIC, timeout of info functions

    def __init__(self, port: str, verb_freq: int = 1) -> None:
        self.__init_serial(port)
        self.__init_thread()
        self.__init_data_structures()

        time.sleep(10e-3)
        self._ser.reset_input_buffer()
        self.start()  # starts the thread

        self.__init_get_device_info()

        self.set_verbose_freq(verb_freq)

    def close(self, blocking: bool = True):
        self.STOP_EVENT.set()
        if blocking:
            self.join()

    def __init_thread(self):
        Thread.__init__(self)
        self.name = "naneos-partector2-serial"
        self.STOP_EVENT = Event()

    def __init_data_structures(self):
        self._data = Queue()
        self._info = Queue()

    def __init_serial(self, port: str):
        self._ser = serial.Serial(port=port, baudrate=9600, timeout=self.TIMEOUT_SER)
        self.set_verbose_freq(0)

    def __init_get_device_info(self):
        try:
            self._serial_number = self.get_serial_number_secure()
            self._firmware_version = self.get_firmware_version()
        except Exception:
            port = self._ser.port
            self.close()
            raise ConnectionError(f"No partector2 on port {port}.")

    def run(self):
        while not self.STOP_EVENT.is_set():
            try:
                self.__serial_reading_routine()
            except Exception as e:
                if not self.__check_device_connection():
                    self.close(blocking=False)
                    p = self._ser.port
                    self._ser.close()
                    raise Exception(f"P2 on port {p} disconnected! Prev exception: {e}")
                    # TODO: raise CUSTOM exception here

        self._ser.close()

    def __serial_reading_routine(self):
        data = [datetime.now(tz=timezone.utc)] + self._readline().split("\t")
        self._data.put(data) if len(data) == len(
            PARTECTOR2_DATA_STRUCTURE
        ) else self._info.put(data)

    def __check_device_connection(self):
        """Checks if P2 is still connected!"""
        if self.STOP_EVENT.is_set():
            return False

        try:
            sn = self.get_serial_number_secure()
            if sn != self._serial_number:
                raise Exception(f"SN changed from {self._serial_number} to {sn}")
        except Exception:
            return False

        return True

    # serial methods
    def __check_serial_connection(self):
        """Tries to reopen a closed connection. Raises exceptions on failure."""
        for _ in range(3):
            self._ser.open() if not self._ser.isOpen() else None
            if self._ser.isOpen():
                return None
        raise Exception("Was not able to open the Serial connection.")

    def _write(self, line: str):
        self.__check_serial_connection()
        self._ser.write(line.encode())

    def _readline(self) -> str:
        self.__check_serial_connection()
        data = self._ser.readline().decode()
        return data.replace("\r", "").replace("\n", "").replace("\x00", "")

    # data methods
    def clear_data_cache(self):
        self._data.queue.clear()

    def get_data_list(self) -> list:
        """Returns the cache as list with timestamp as first element."""
        data_casted = []
        data = list(self._data.queue)
        self.clear_data_cache()

        for line in data:
            try:
                data_casted.append(self.__cast_splitted_input_string(line))
            except Exception as excep:
                print(f"Exception during casting (sp): {excep}")

        return data_casted

    def get_lambda_upload_list(self, data=None) -> list:
        if not data:
            data = self.get_data_list()
        return _get_lambda_upload_list_serial(data, self._serial_number)

    def get_data_pandas(self, data=None) -> pandas.DataFrame:
        if not data:
            data = self.get_data_list()

        columns = PARTECTOR2_DATA_STRUCTURE.keys()
        df = pandas.DataFrame(data, columns=columns).set_index("dateTime")
        return df

    def __cast_splitted_input_string(self, line: str):
        for i, data_type in enumerate(PARTECTOR2_DATA_STRUCTURE.values()):
            if type(line[i]) is not data_type:
                line[i] = data_type(line[i])

        return line

    # user wrappers
    def __user_wrapper(self, func):
        """Wraps user func in try-except block. Forwards exceptions to the user."""
        for _ in range(self.SERIAL_RETRIES):
            try:
                return func()
            except Exception as e:
                excep = f"Exception occured during user function call: {e}"
        raise Exception(excep)

    def get_serial_number(self) -> int:
        return self.__user_wrapper(self.__get_serial_number)

    def get_serial_number_secure(self) -> int:
        for _ in range(3):
            serial_numbers = [self.get_serial_number() for _ in range(3)]
            if all(x == serial_numbers[0] for x in serial_numbers):
                return serial_numbers[0]
        raise Exception("Was not able to fetch the serial number (secure)!")

    def get_firmware_version(self) -> str:
        return self.__user_wrapper(self.__get_firmware_version)

    def set_verbose_freq(self, freq: int):
        self.__set_verbose_freq(freq)

    # user protected methods
    def __get_and_check_info(self, leng: int = 2) -> list:
        data = self._info.get(timeout=self.TIMEOUT_INFO)
        if len(data) != 2:
            raise Exception(f"Info length {len(data)} not matching {leng}: {data}")
        return data

    def __get_serial_number(self) -> int:
        self._info.queue.clear()
        self._write("N?")
        return int(self.__get_and_check_info()[1])

    def __get_firmware_version(self) -> str:
        self._info.queue.clear()
        self._write("f?")
        return int(self.__get_and_check_info()[1])

    # TODO: check if frequency was set the right way
    # TODO: Ask Martin for an X? option in the protocol
    def __set_verbose_freq(self, freq: int):
        self._write(f"X000{freq}!")



import time

from naneos.partector2 import Partector2, scan_for_serial_partector2

# make sure a physical partector2 is connected


def test_port_scanning():
    ports = scan_for_serial_partector2()
    assert len(ports) > 0
    assert isinstance(ports, dict)
    assert isinstance(list(ports.keys())[0], int)
    assert isinstance(list(ports.values())[0], str)


def test_partector_data_list():
    p2 = _find_and_connect_partector2(2)

    time.sleep(2)

    data = p2.get_data_list()

    assert 10 < len(data) < 30

    p2.close(blocking=True)


def test_partector_pandas():
    p2 = _find_and_connect_partector2(3)

    time.sleep(2)

    data = p2.get_data_pandas()

    assert 100 < len(data) < 300

    p2.close(blocking=True)


def test_partector_upload_format():
    p2 = _find_and_connect_partector2(2)
    time.sleep(1)

    data = p2.get_lambda_upload_list()
    assert 5 < len(data) < 15
    assert type(data) == list
    assert type(data[0]) == dict

    p2.close(blocking=True)


def test_partector_freq_change():
    p2 = _find_and_connect_partector2(0)
    time.sleep(2)

    data = p2.get_data_pandas()
    assert len(data) == 0

    p2.set_verbose_freq(1)
    time.sleep(2)
    data = p2.get_data_pandas()
    assert 1 <= len(data) <= 3

    p2.set_verbose_freq(2)
    time.sleep(2)
    data = p2.get_data_pandas()
    assert 10 < len(data) < 30

    p2.set_verbose_freq(3)
    time.sleep(2)
    data = p2.get_data_pandas()
    assert 100 < len(data) < 300

    p2.close(blocking=True)


def _find_and_connect_partector2(freq: int = 0):
    scan = scan_for_serial_partector2()
    sn = list(scan.keys())[0]
    port = list(scan.values())[0]

    p2 = Partector2(port, verb_freq=freq)
    assert p2._serial_number == sn

    return p2



import glob
import sys

import serial


def list_serial_ports():
    """Lists serial port names

    :raises EnvironmentError:
        On unsupported or unknown platforms
    :returns:
        A list of the serial ports available on the system
    """
    if sys.platform.startswith("win"):
        ports = ["COM%s" % (i + 1) for i in range(256)]
    elif sys.platform.startswith("linux") or sys.platform.startswith("cygwin"):
        # this excludes your current terminal "/dev/tty"
        ports = glob.glob("/dev/tty[A-Za-z]*")
    elif sys.platform.startswith("darwin"):  # mac
        ports = glob.glob("/dev/tty.*")
        ports = [port for port in ports if "Bluetooth" not in port]
    else:
        raise EnvironmentError("Unsupported platform")

    result = []

    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result



from math import pi

from naneos.partector2._data_structure import get_p2_idx as get_i


def _get_lambda_upload_list_serial(data_list: list, serial_number: int) -> list:
    upload_list = []

    for entry in data_list:
        mass = __calc_particle_mass(entry[get_i("number")], entry[get_i("diameter")])
        upload_dict = {
            "deviceName": "P2",
            "timestamp": int(entry[get_i("dateTime")].timestamp()),
            "locationData": {"latitude": 0.0, "longitude": 0.0},  # TODO: implement
            "partectorData": {
                "average_particle_diameter": entry[get_i("diameter")],
                "battery_voltage": entry[get_i("batt_voltage")],
                "device_status": entry[get_i("device_status")],
                "ldsa": entry[get_i("LDSA")],
                "particle_mass": mass,
                "particle_number_concentration": entry[get_i("number")],
                "relative_humidity": entry[get_i("RHcorr")],
                "serial_number": serial_number,
                "temperature": entry[get_i("T")],
            },
        }
        upload_list.append(upload_dict)

    return upload_list


def __calc_particle_mass(number, diameter):
    """Calculates the particle_mass from number and diameter.
    Due to Martin the obtained values are very uncertain."""
    return number * pi * 6.38 / 6.0 * pow(diameter, 3) * 1.2e-9



import time

from naneos.partector2_ble import Partector2Ble


def test_bluetooth_connection():
    p2 = Partector2Ble()
    time.sleep(2)
    assert p2.get_and_clear_results() != {}
    p2.stop_scanning()



def _get_lambda_upload_list_ble(data_dict: dict, sn_exclude: list = []) -> list:
    """
    Returns a list of dicts with the data to upload to lambda.
    :param data_dict:
    :param sn_exclude:
    :return:
    """
    data_list = []
    for sn, data in data_dict.items():
        if sn in sn_exclude:
            continue

        for entry in data:
            upload_dict = {
                "deviceName": "P2",
                "timestamp": int(entry["dateTime"].timestamp()),
                # TODO: implement
                "locationData": {"latitude": 0.0, "longitude": 0.0},
                "partectorData": {
                    "average_particle_diameter": entry["diameter"],
                    "battery_voltage": entry["batt_voltage"],
                    "device_status": entry["device_status"],
                    "ldsa": entry["LDSA"],
                    "particle_mass": entry["particle_mass"],
                    "particle_number_concentration": entry["number"],
                    "relative_humidity": entry["RHcorr"],
                    "serial_number": sn,
                    "temperature": entry["T"],
                },
            }
            data_list.append(upload_dict)

    return data_list
    # data_dict structure: structure =
    # {8000: [["data_list"], ["data_list"]], 8001: [["data_list"], ["data_list"]]}
    # serial numbers exclude: when fetching from serial and ble, prefer serial data



import time

from naneos.partector2 import Partector2, scan_for_serial_partector2

# lists all available Partector2 devices
x = scan_for_serial_partector2()

# connect to the first device
myP2 = Partector2(list(x.values())[0], 1)
time.sleep(2)

# get the data as a pandas dataframe
data = myP2.get_data_pandas()
print(data)

myP2.close()



import datetime

# v295
PARTECTOR2_DATA_STRUCTURE = {
    "dateTime": datetime.datetime,
    "runtime_min": float,
    "idiff_global": float,
    "ucor_global": int,
    "hiresADC1": float,
    "hiresADC2": float,
    "EM_amplitude1": float,
    "EM_amplitude2": float,
    "T": float,
    "RHcorr": int,
    "device_status": int,
    "deposition_voltage": int,
    "batt_voltage": float,
    "flow_from_dp": float,
    "LDSA": float,
    "diameter": float,
    "number": int,
    "dP": int,
    "P_average": float,
}


def get_p2_idx(key: str) -> int:
    return list(PARTECTOR2_DATA_STRUCTURE.keys()).index(key)
